import os
import math
import re

outDataPath = os.path.join("requiredFiles", "")
MY_MAX = 10000
MY_MIN = -10000




################################################################################
# Functions for fasta formatted file parsing and validation
################################################################################
def isFasta(fileInLines):
    '''
    Input: List of lines of a stream. A line is defined as a sequence of characters that does not contain 
    newline character anywhere but at the very end.
    
    Function: Checks if the given stream is a valid fasta file. Valid fasta file contains 
    a number of protein sequences that could extend to more than 1 line separated by description lines
    that start with '>' character. Description line should come before corresponding protein sequence.
    
    Output: Either a non-empty string containing the error message or an empty string indicating successful 
    validation is returned.
    '''
    DESCRIPTION = 1
    SEQUENCE = 2
    state = 0 # state variable is initialized to zero and will be taking values DESCRIPTION and SEQUENCE during the processing of the given stream
    
    # Validation of the given stream to make sure there are no missing description lines or 
    # corresponding sequences, and these are in proper, alternating order 
    for line in fileInLines:
        if(line == ""): #ignore blank lines
            pass
        elif(line[0] == ">"): #Then, this line must be a description line
            if state == DESCRIPTION:
                return "Not a fasta file: Missing sequence"    
            state = DESCRIPTION
        else: #Then, this line must be part of a sequence
            if state == SEQUENCE or state == DESCRIPTION:
                state = SEQUENCE
            else:
                return "Not a fasta file: Missing description line"
    
    if state != SEQUENCE:
        return "Not a fasta file: Missing sequence"
        
    return ""
 
def extractIdFromDescriptionLine(line):
    '''
    Input: A fasta description line that starts with the character '>'
    
    Function: Assuming either of the following two scenarios, extracts id from the given line.
              - id follows '>' character in a line and either it is followed by nothing or it
              is followed by '|' character marking the end of the id and the beginning of the
              protein description (which will be ignored).
              - line starts with the character sequence '>gi|' and id follows this, '|' character 
              marking the end of the id
              
    Output: Returns the extracted id as a string          
    '''
    giPattern = re.compile('>gi|')
    #extract protein id from the description line
    matchObj = re.search('>gi\\|', line)
    if(matchObj is not None):
        line = line[matchObj.end():]
        record = line.split("|"); 
        return record[0].strip()
    elif len( line[1:].split('|')) > 1:
        return line[1:].split('|')[0].strip()
    elif len( line[1:].split()) > 1:
        return line[1:].split()[0].strip()
    else:
        return line[1:].strip()
 
 
def validify(seq):  
    aaList =   ['A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G', 'H', 'I', 'L', 'K', 'M', 'F', 'P', 'S', 'T', 'W', 'Y', 'V']
    aaList_lo = [el.lower() for el in  aaList]
    acceptable = ''.join(aaList + aaList_lo)
    validSeq = seq
    for ch in seq:
        if not ch in acceptable:
            validSeq = ''.join(validSeq.split(ch))
    return validSeq
         
def fetchIDandSequence(fastaInLines):
    ''' 
    Input: List of lines of a stream containing a valid fasta formatted file. Assumes a valid fasta formatted 
    file is given as input (check isFasta function above to learn what valid fasta format is supposed to be)
    
    Function: Parses lines of a fasta formated file to get the list of IDs and sequences of proteins
    
    Output: Parallel lists of ids and protein sequence contained in the given stream
    '''
    DESCRIPTION = 1
    SEQUENCE = 2
    state = 0 # state variable is initialized to zero and will be taking values DESCRIPTION and SEQUENCE during the processing of the given stream
    idList = []
    sequences=[]
    seq = ""  
    for line in fastaInLines:
        if(line == ""): #ignore blank lines
            pass
        elif(line[0] == ">"): #Then, this line must be a description line        
            idList.append(extractIdFromDescriptionLine(line))   
            if state == SEQUENCE:
                sequences.append(validify(seq))
            state = DESCRIPTION
        else: #Then, this line must be part of a sequence
            if (state == DESCRIPTION): # beginning of a new sequence
                state = SEQUENCE
                seq = ""  
            seq += line.strip()
    sequences.append(validify(seq)) #Add last sequence to the list
    return sequences, idList

class svmParameters:
    def __init__(self, c, g, meme_c, meme_g):
        self.c = c
        self.g = g
        self.meme_c = meme_c
        self.meme_g = meme_g
        
def copyLists(given):
    newList = []
    for i in range(len(given)):
        tmp = []
        for j in range(len(given[i])):
            tmp.append(given[i][j])
        newList.append(tmp)
    return newList

def read1DFile(fileName):
    ''' Reads a file which contains an item per line until the end of file or until an empty line '''
    holdingList = []
    fileHandle = open(fileName,"r")
    for line in fileHandle:
        line = line.rstrip()
        if(line == ""):
            break
        else:
            holdingList.append(line)
    fileHandle.close()
    return holdingList

class posNegPair:
    
    def __init__(self):
        self.pos = 0.0
        self.neg = 0.0
        
    def add(self, pair):
        self.pos += pair.pos
        self.neg += pair.neg
        
    def divide(self, pair):
        self.pos /= pair.pos
        self.neg /= pair.neg
        
    def increment(self, isPos):
        if isPos:
            self.pos += 1
        else:
            self.neg += 1

class normalizationData:

    def __init__(self):
        self.mapForDelete = []
    def initiliaze(self, featureCount):
        self.featureCount = featureCount
        self.maxOverFeatures = []
        self.minOverFeatures = []
        self.totalOverFeatures = []
        self.powerOverFeatures = []
        self.stdevOverFeatures = []
        self.countOverFeatures = []
        self.avgOverFeatures = []
        self.mapForDelete = []
        for i in range(featureCount):
            self.totalOverFeatures.append(0.0)
            self.powerOverFeatures.append(0.0)
            self.minOverFeatures.append(MY_MAX)
            self.maxOverFeatures.append(MY_MIN)    
            self.stdevOverFeatures.append(0.0) 
            self.countOverFeatures.append(0)
            self.avgOverFeatures.append(0.0) 
            self.mapForDelete.append(0) 
            
    def submitInstance(self,featureID, value):
        self.totalOverFeatures[featureID] += value
        self.powerOverFeatures[featureID] += pow(value, 2)
        self.minOverFeatures[featureID] = min(self.minOverFeatures[featureID], value)    
        self.maxOverFeatures[featureID] = max(self.maxOverFeatures[featureID], value) 
        self.countOverFeatures[featureID] += 1
        self.stdevOverFeatures[featureID] = 0.0 #forces recalculation
        self.avgOverFeatures[featureID] = 0.0 #forces recalculation
        
    def getAvg(self,featureID):
        if (self.avgOverFeatures[featureID] == 0.0):
            if (self.countOverFeatures[featureID] != 0):
                self.avgOverFeatures[featureID] = self.totalOverFeatures[featureID] / self.countOverFeatures[featureID]
            else:
                self.avgOverFeatures[featureID] = self.totalOverFeatures[featureID]
        return self.avgOverFeatures[featureID]; 
    
    def getPower(self,featureID):
        if (self.powerOverFeatures[featureID] == 0.0):
            if (self.countOverFeatures[featureID] != 0):
                self.powerOverFeatures[featureID] = self.powerOverFeatures[featureID] / self.countOverFeatures[featureID]
            else:
                self.powerOverFeatures[featureID] = self.powerOverFeatures[featureID]
        return self.powerOverFeatures[featureID]; 
    
    def getStdev(self,featureID):
        if (self.stdevOverFeatures[featureID] == 0.0):
            self.stdevOverFeatures[featureID] = math.sqrt( self.getPower(featureID) - pow(self.getAvg(featureID), 2) )
        return self.stdevOverFeatures[featureID]; 

    def fillMapForDelete(self):
        for i in range(self.featureCount):
            if(self.getAvg(i) == 0):
                self.mapForDelete[i] = 1
                
    def writeToFile(self, outfile):
        outfile = open(outfile, 'w')
        for el in self.maxOverFeatures:
            outfile.write(str(el) + "\t")
        outfile.write('\n')
        for el in self.minOverFeatures:
            outfile.write(str(el) + "\t")
        outfile.write('\n')
        outfile.close()
    
    def readFromFile(self, infile):
        infile = open(infile, 'r')
        maxLine = infile.readline()
        record = maxLine.split('\t')
        self.maxOverFeatures = []
        for el in record:
            el = el.strip()
            if el:
                self.maxOverFeatures.append(float(el.strip()))
            
        minLine = infile.readline()
        record = minLine.split('\t')
        self.minOverFeatures = []
        for el in record:
            el = el.strip()
            if el:
                self.minOverFeatures.append(float(el.strip()))
        infile.close()
        
        self.mapForDelete = []
        for i in range(len(self.maxOverFeatures)):
            self.mapForDelete.append(0)

class normalizationHub:
    
    def __init__(self):
        pass
    
    def initializeFull(self, pos_fv, neg_fv, posTest_fv, negTest_fv):
        normData =   self.normalizeParameterCalculate  (pos_fv, neg_fv)
        pos_fv, neg_fv = self.normalizeMinMax(pos_fv, neg_fv, normData) #normalizeGaussian  normalizeMinMax
        posTest_fv, negTest_fv = self.normalizeMinMax(posTest_fv, negTest_fv, normData) #normalizeGaussian  normalizeMinMax
        return normData
        
    def initializeTest(self, fv, normDataFile):
        normData = normalizationData()
        normData.readFromFile(normDataFile)
        fv = self.normalizeMinMax_helper(fv, normData) 
    
     
    def normalizeParameterCalculate(self, fvList1, fvList2):
        
        featureCount = len(fvList1[0]) #assuming other list will have same number of features
        normData = normalizationData()
        normData.initiliaze(featureCount)
    
        for j in range(featureCount): #for each feature
            for fvList in [fvList1, fvList2]:
                for i in range(len(fvList)):
                    normData.submitInstance(j, fvList[i][j])
       
        normData.fillMapForDelete()
        return  normData 
                   
    
    def normalizeMinMax_helper(self, fvList1, normData):      
        featureCount = len(fvList1[0]) #assuming other list will have same number of features           
        for j in range(featureCount): #for each feature
            for i in range(len(fvList1)):
                diff = (normData.maxOverFeatures[j] - normData.minOverFeatures[j])
                if diff:
                    fvList1[i][j]  = (fvList1[i][j] - normData.minOverFeatures[j]) / diff
     
        featureCount = len(fvList1[0]) #assuming other list will have same number of features   
        newfvList1 = []    
        for i in range(len(fvList1)):
            tmp = []
            for j in range(featureCount): #for each feature
                if not normData.mapForDelete[j]:
                    tmp.append(fvList1[i][j])
            newfvList1.append(tmp)          
        return newfvList1
        
    def normalizeMinMax(self, fvList1, fvList2, normData):      
        return self.normalizeMinMax_helper(fvList1, normData), self.normalizeMinMax_helper(fvList2, normData)

                     
    def normalizeGaussian(self, fvList1, fvList2, normData): 
        
        featureCount = len(fvList1[0]) #assuming other list will have same number of features
        
        for j in range(featureCount): #for each feature
            for fvList in [fvList1, fvList2]:
                for i in range(len(fvList)):
                    diff = normData.getStdev(j) 
                    if diff:
                        fvList[i][j]  = (fvList[i][j] - normData.getAvg(j)) / diff 
    
class svmClassifier:
    def __init__(self, whereIsSVM, tempFolderPath):
        self.whereIsSVM = whereIsSVM   
        self.tempFolderPath = tempFolderPath
        self.index = -1
     
    def binaryFeatures_specialTreatment(self):
        
        maxValue = 1
        minValue = 0.1
        for i in range(len(self.pos_fv)):
            for j in range(len(self.pos_fv[i])):
                if ( self.pos_fv[i][j] == 0):
                    self.pos_fv[i][j] = minValue;
                else:
                    self.pos_fv[i][j] = maxValue;
                    
        for i in range(len(self.neg_fv)):
            for j in range(len(self.neg_fv[i])):
                if ( self.neg_fv[i][j] == 0):
                    self.neg_fv[i][j] = minValue;
                else:
                    self.neg_fv[i][j] = maxValue;
                    
        for i in range(len(self.posTest_fv)):
            for j in range(len(self.posTest_fv[i])):
                if ( self.posTest_fv[i][j] == 0):
                    self.posTest_fv[i][j] = minValue;
                else:
                    self.posTest_fv[i][j] = maxValue;
                    
        for i in range(len(self.negTest_fv)):
            for j in range(len(self.negTest_fv[i])):
                if ( self.negTest_fv[i][j] == 0):
                    self.negTest_fv[i][j] = minValue;
                else:
                    self.negTest_fv[i][j] = maxValue;
             
    def runTrain(self, parameters, pos_fv, neg_fv, posTest_fv, negTest_fv, normDataFile, givenModelFileName = ""):
        self.parameters = parameters 
        #Test on itself  
        posTest_fv = pos_fv 
        negTest_fv = neg_fv
        
        #copy data before normalization
        self.pos_fv = copyLists(pos_fv)
        self.neg_fv = copyLists(neg_fv)
        self.posTest_fv = copyLists(posTest_fv)
        self.negTest_fv = copyLists(negTest_fv)  
        
        if (self.index == -1):
            postId = ""
        else:
            postId = str(self.index)
            self.index += 1;
            
        trainOut  = self.tempFolderPath + "svm_trainOut" + postId
        testOut  = self.tempFolderPath + "svm_testOut" + postId
        testPredictionFileName = self.tempFolderPath + "svm_testPrediction" + postId
        testFileName = self.tempFolderPath + "svm_testFile" + postId
        trainFileName = self.tempFolderPath + "svm_trainFile" + postId 
        
        for fileX in [testFileName, trainFileName, trainOut, testOut, testPredictionFileName]:
            if os.path.isfile(fileX):
                os.remove(fileX)
    
        #normalize
        constructionDoesItAll = normalizationHub()
        normData = constructionDoesItAll.initializeFull(self.pos_fv, self.neg_fv, self.posTest_fv, self.negTest_fv)
        normData.writeToFile(normDataFile)
        #file prepare
        self.svmFilePrepare(trainFileName, self.pos_fv, self.neg_fv)
        self.svmFilePrepare(testFileName, self.posTest_fv, self.negTest_fv) 
        
        c = parameters.c
        g = parameters.g
        
        ###Training
        #cost = len(self.negTest_fv)/float(len(self.posTest_fv)) #-j "+str(cost)+"
        command1= self.whereIsSVM + "svm_learn -v 1 -t 2 -g "+str(g)+ " -c "+str(c)+" " + \
                  trainFileName+ " " + givenModelFileName + " > " + trainOut 
        os.system(command1); #print(command1)
        ###Testing            
        command2 = self.whereIsSVM + "svm_classify -f 1 " + testFileName + " " + givenModelFileName + " " + \
                   testPredictionFileName + "> " + testOut  
        os.system(command2) ; #print(command2)

    
        #accuracy, x, y  = parseSvmClassificationOutput(parameters.classifierTestOutputFileName)
        predictions = self.svmPredictionOutputParse(testPredictionFileName)      
        #print(predictions)
        
        return predictions
    
    def runTest(self, parameters, fv, normDataFile, givenModelFileName = ""):
        self.parameters = parameters 

        #copy data before normalization
        self.fv = copyLists(fv)
           
        testOut  = self.tempFolderPath + "svm_testOut" 
        testPredictionFileName = self.tempFolderPath + "svm_testPrediction" 
        testFileName = self.tempFolderPath + "svm_testFile" 
              
        for fileX in [testFileName, testOut, testPredictionFileName]:
            if os.path.isfile(fileX):
                os.remove(fileX)
    
        #normalize
        constructionDoesItAll = normalizationHub()
        normData = constructionDoesItAll.initializeTest(self.fv, normDataFile)
        #file prepare
        self.svmFilePrepareForTest(testFileName, self.fv) 
 
        ###Testing            
        command2 = self.whereIsSVM + "svm_classify -f 1 " + testFileName + " " + givenModelFileName + " " + \
                   testPredictionFileName + "> " + testOut  
        os.system(command2) ; #print(command2)
        #accuracy, x, y  = parseSvmClassificationOutput(parameters.classifierTestOutputFileName)
        predictions = self.svmPredictionOutputParse(testPredictionFileName)      
        #print(predictions)     
        return predictions
    
    def svmFilePrepareHelper(self, svmFile,data,classTag):
        for i in range(len(data)):
            svmFile.write(classTag + "")
            count = 1
            for j in range(len(data[i])):
                feature = "%d:" %count
                value = "%.2f " %float(data[i][j])
                svmFile.write(feature + value)
                count += 1
            svmFile.write("\n")
        
    def svmFilePrepare(self, svmFileName,positiveData,negativeData):
        
        svmFile = open(svmFileName,"w")
        self.svmFilePrepareHelper(svmFile,positiveData,"+1 ")
        self.svmFilePrepareHelper(svmFile,negativeData,"-1 ")
        svmFile.close()
        
    def svmFilePrepareForTest(self, svmFileName,data):
        svmFile = open(svmFileName,"w")
        self.svmFilePrepareHelper(svmFile,data,"+1 ")
        svmFile.close()
        
    def svmPredictionOutputParse(self,testPredictionFileName):
        predictions = read1DFile(testPredictionFileName)
        predictionList = []
        for i in range(len(predictions)):
            p = predictions[i]
            if(float(p) > 0):
                predictionList.append(1)
            else:
                predictionList.append(0)
        #print(predictionList)
        return predictionList
  
class GO_based_classification:
    
    def __init__(self, reqFilePath):
        self.reqFilePath = reqFilePath
        self.allGoMap = {}
        self._goMapAllRead()
        
    def _goMapAllRead(self):
        '''
        Parses a file containing lines in the following format: 
        <protein id><list of go term ids associated with the protein>
        '''
        goMapFile = self.reqFilePath + "ampAll.goMap"
        goMapFile = open(goMapFile, 'r')
        for line in goMapFile:
            record = line.strip().split('\t')
            self.allGoMap[record[0]] = []
            for el in record[1:]:
                self.allGoMap[record[0]].append(int(el))
            
    def _goTermFeatureMapPrepare(self, posTrainIds, negTrainIds):
        '''
        Calculate the probabilities of go terms appearing in the given 
        positive and negative training sets 
        goScoreMap associates go terms (that appeared in one of the given 
        training sets) with the 2 probability values
        '''
        goScoreMap = {}
        
        idListList = [posTrainIds, negTrainIds]
        isPosFlags = [1, 0]
        
        for idx, idList in enumerate(idListList):
            isPos = isPosFlags[idx]
            for proteinId in idList:
                for el in self.allGoMap[proteinId]:
                    if not el in goScoreMap:
                        goScoreMap[el] = posNegPair()
                    #update entries   
                    goScoreMap[el].increment(isPos)
        
        count = posNegPair()
        for el in goScoreMap:
            count.add(goScoreMap[el])
        
        for el in goScoreMap:
            goScoreMap[el].divide(count);
                                          
        return goScoreMap
           
    def printGoScoreMap(self, goScoreMap):  
        tmp = {} 
        for el in goScoreMap:
            tmp[el] = [goScoreMap[el].pos,goScoreMap[el].neg]     
        print(tmp)
                       
    def _goTermBasedClassify(self, idx, goScoreMap):  
        scores = posNegPair() 
        for el in self.allGoMap[idx]: #for all go terms associated with the given id
            if el in goScoreMap: #check out the prob values for being positive or negative
                scores.pos += goScoreMap[el].pos
                scores.neg += goScoreMap[el].neg  #TRIED math.log(prob) at some point (check if prob is not 0) 
        return scores

    def _goTermBasedClassifyBatch(self, ids, goScoreMap):
        scores = []
        for idx in ids:
            scores.append(self._goTermBasedClassify(idx, goScoreMap))
        return scores
    
    def goClassifier(self, idListP_train, idListN_train, idListP, idListN):
        
        #prepare map, by only looking at train
        goScoreMap = self._goTermFeatureMapPrepare(idListP_train, idListN_train)
        
        #self.printGoScoreMap(goScoreMap)
        
        #now calculate predictions on test data
        posTestGoScores = self._goTermBasedClassifyBatch(idListP, goScoreMap)
        negTestGoScores = self._goTermBasedClassifyBatch(idListN, goScoreMap)
                                
        predictions = []        
        testScoreListList = [posTestGoScores, negTestGoScores]
        isPosFlags = [1, 0]
        
        truePredCounts = posNegPair()
        for idx, testScoreList in enumerate(testScoreListList):
            for testScore in testScoreList:
                if testScore.pos > testScore.neg:
                    predictions.append(1)
                    if isPosFlags[idx]:
                        truePredCounts.pos += 1
                elif testScore.pos == testScore.neg:
                    predictions.append(-1)
                else:
                    predictions.append(0)
                    if not isPosFlags[idx]:
                        truePredCounts.neg += 1

        #accuracy = (truePredCounts.pos + truePredCounts.neg) / float(len(posTestGoScores) + len(negTestGoScores))
        return predictions

def mergeVectorRepr(lists):
    mainList = []
    for i in range(len(lists)):   #len(lists) gives number of feature representations to merge
        for el in lists[i]:
            mainList.append(el)
    return mainList

def readGESindex(gesFile):
    ''''read GES index (Goldman-Engelman-Steitz (GES) hydropathy scale)'''
    gesFile = open(gesFile,'r')
    GES = {}
    for line in gesFile:
        r = line.strip().split('\t')
        GES[r[0].strip()] = float(r[1].strip())
    gesFile.close()
    return GES
#######################################################

def writeSequenceInFastaFormat(sequence, fileHandle):
    for i in range(len(sequence)):
        fileHandle.write(sequence[i])
    
def createFastaFile(ids, sequences, outFileName):
    outFile = open (outFileName, "w")
    for i in range(len(ids)):
        outFile.write(">" + ids[i] + "\n")
        writeSequenceInFastaFormat(sequences[i], outFile)
        outFile.write("\n")
    outFile.close()
    
def fastaFileRead (fname, idSep):
    # Read in protein data file and create lists for annotation
    # and amino acid sequences. 
    id_list = []
    aa_list = []
    inProteinFlag = 0
    
    file = open(fname, 'r')
    for line in file:
        if len(line) == 0:
            pass
        elif line[0] == '>':  #annotation line
            
            id_list.append(line[1:].rstrip())
            inProteinFlag = 0
        elif inProteinFlag == 0:
            aa_list.append(line.rstrip().upper())
            #aa_list.append("".join(line.rstrip().upper().split('X')))
            inProteinFlag = 1
        elif inProteinFlag == 1:
            aa_list[len(aa_list) - 1] = aa_list[len(aa_list) - 1] + line.rstrip().upper() #"".join(line.rstrip().upper().split('X'))

    #parse id list to get gi list
    gi_list = []
    for item in id_list:
        if idSep:
            gi_list.append(item.split(idSep)[0])
        else:
            gi_list.append(item)
    return gi_list, aa_list
    
